"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
from jax.errors import TracerArrayConversionError

from qrisp.alg_primitives.state_preparation.qiskit_state_preparation import (
    prepare_qiskit,
)
from qrisp.alg_primitives.state_preparation.qswitch_state_preparation import (
    prepare_qswitch,
)
from qrisp.jasp.tracing_logic import check_for_tracing_mode


def prepare(qv, target_array, reversed: bool = False, method: str = "auto"):
    r"""
    Prepare a quantum state on ``qv`` from a target amplitude vector.

    Given a vector :math:`b=(b_0,\dotsc,b_{N-1})` (corresponding to ``target_array``),
    this routine prepares the quantum state:

    .. math::

        \sum_{i=0}^{N-1} b_i \ket{i}.

    The ``target_array`` input vector is normalized internally, i.e.

    .. math::

        \tilde b_i = \frac{b_i}{\|b\|},
        \qquad
        \ket{0} \;\mapsto\; \sum_{i=0}^{N-1} \tilde b_i \ket{i}.

    By default, the little-endian convention is used for the computational basis ordering.
    For example, for a 2-qubit system, the basis states are ordered as

    .. math::

        \ket{0} \equiv \ket{q_0 = 0, q_1 = 0}, \quad \ket{1} \equiv \ket{q_0 = 1, q_1 = 0}, \dotsc


    Parameters
    ----------
    qv : QuantumVariable
        Quantum variable to prepare.

    target_array : numpy.ndarray or jax.numpy.ndarray
        Target amplitude vector :math:`b`. Must have length :math:`2^n` where
        :math:`n` is the size of ``qv`` (validated for concrete arrays).

    reversed : bool, optional
        If ``True``, applies a bit-reversal permutation (big-endian ordering).
        Default is ``False``.

    method : {'qiskit', 'qswitch', 'auto'}, optional
        Compilation method for state preparation. Possible values are:

        - ``'qiskit'``: requires concrete arrays (e.g. NumPy).
        - ``'qswitch'``: supports traced arrays (e.g. JAX tracers in Jasp mode).
          Note that shape validation is not performed in Jasp mode.
        - ``'auto'``: automatically selects between the above.

        Default is ``'auto'``.

    Examples
    --------

    In this example, we create a :ref:`QuantumFloat` and prepare the normalized state
    $\sum_{i=0}^3 \tilde b_i\ket{i}$ for $\tilde b=(0,1,2,3)/\sqrt{14}$.

    ::

        import numpy as np
        from qrisp import QuantumFloat, prepare

        b = np.array([0, 1, 2, 3], dtype=float)
        b /= np.linalg.norm(b)

        qf = QuantumFloat(2)
        prepare(qf, b)

    We can verify that the state has been correctly prepared.

    For example, we can use the ``statevector`` method to get a function that maps basis states to amplitudes:

    ::

        sv_function = qf.qs.statevector("function")

        print(f"b[1]: {b[1]:.6f} -> {sv_function({qf: 1}):.6f}")
        # b[1]: 0.267261 -> 0.267261-0.000000j
        print(f"b[2]: {b[2]:.6f} -> {sv_function({qf: 2}):.6f}")
        # b[2]: 0.534522 -> 0.534522-0.000000j

    where index 1 in little-endian corresponds to the basis state :math:`\ket{q_0=1, q_1=0}`
    and index 2 to :math:`\ket{q_0=0, q_1=1}`.  With ``reversed=True``, we can
    switch to big-endian ordering. That is, we can map ``b[1]`` to ``sv_function({qf: 2})``
    instead, and so on.

    We can perform a similar verification even if the statevector is not directly
    accessible (for example when running on hardware), by using measurement results:

    ::

        qf = QuantumFloat(2)
        prepare(qf, b)

        res_dict = qf.get_measurement()

        ref = np.sqrt(res_dict[1])
        amps = {k: round(np.sqrt(v) / ref) for k, v in res_dict.items()}

        print(amps)
        # Yields: {3: 3, 2: 2, 1: 1}

    The output indicates that the magnitudes of the amplitudes for the basis states
    :math:`\ket{1}`, :math:`\ket{2}`, and :math:`\ket{3}` are in the ratio :math:`1 : 2 : 3`,
    exactly matching the input vector :math:`b = (0,1,2,3)` up to normalization.

    .. note::

        This primitive is not yet compatible with QuantumEnvironments
        (e.g. ``invert`` or ``control``) in Jasp mode when using the ``qswitch`` method.
        Trying to use it within such environments, for example by writing:

        ::

            from qrisp.jasp.evaluation_tools import terminal_sampling

            @terminal_sampling
            def circuit():

                (...)

                with invert():
                    prepare(..., method="qswitch")

        currently results in an error.

        Furthermore, it is currently not possible to prepare
        a state with 64 or more qubits using the ``qswitch`` method.

    """

    if method not in {"auto", "qiskit", "qswitch"}:
        raise ValueError("method must be 'auto', 'qiskit', or 'qswitch'")

    is_tracing = check_for_tracing_mode()

    if not is_tracing:
        expected = 1 << qv.size
        if target_array.size != expected:
            raise ValueError(
                f"Statevector length must be {expected} for {qv.size} qubits, "
                f"got {target_array.size}."
            )
        target_array = np.asarray(target_array)
        norm = np.linalg.norm(target_array)
        if np.isclose(norm, 0.0):
            raise ValueError("The provided statevector has zero norm.")
        target_array = target_array / norm

    if method == "auto":
        try:
            target_array = np.asarray(target_array)
            method = "qiskit"
        except TracerArrayConversionError:
            method = "qswitch"

    if method == "qiskit":
        prepare_qiskit(qv, target_array, reversed)
    else:
        prepare_qswitch(qv, target_array, reversed)



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

"""
Automatic detector identification for quantum error correction circuits.

This module provides the :func:`find_detectors` decorator, which leverages
`tqecd <https://github.com/tqec/tqecd>`_ to automatically identify
stim ``DETECTOR`` instructions at Jasp tracing time and emit the
corresponding ``parity()`` calls.

Usage
-----
.. code-block:: python

    from qrisp import *
    from qrisp.jasp.evaluation_tools.stim_extraction import extract_stim
    from qrisp.misc.stim_tools.find_detectors import find_detectors

    @find_detectors
    def syndrome_round(qa):
        reset(qa[1]); reset(qa[3])
        cx(qa[0], qa[1]); cx(qa[2], qa[1])
        cx(qa[2], qa[3]); cx(qa[4], qa[3])
        return measure(qa[1]), measure(qa[3])

    @extract_stim
    def main():
        qa = QuantumArray(qtype=QuantumBool(), shape=(5,))
        detectors, m1, m3 = syndrome_round(qa)
        return detectors, m1, m3
"""

import stim
import numpy as np
import jax
from jax.tree_util import tree_flatten

from qrisp import QuantumArray, QuantumBool
from qrisp.jasp import make_jaspr
from qrisp.jasp.primitives.parity_primitive import parity
from qrisp.circuit import QuantumCircuit as QC

# ───────────────────── stim instruction categories ─────────────────────
_RESET_OPS = frozenset({"R", "RX", "RY"})
_MEASUREMENT_OPS = frozenset({"M", "MX", "MY", "MZ", "MR", "MRX", "MRY"})
_TWO_QUBIT_OPS = frozenset({"CX", "CY", "CZ", "XCZ", "YCZ", "SWAP", "ISWAP"})


# ───────────────────── internal helpers ────────────────────────────────


def _tvals(inst):
    """Return plain target-value list for a stim instruction."""
    return [t.value for t in inst.targets_copy()]


def _build_analysis_args(jaspr, n_qubits, quantum_arrays):
    """Build concrete args for calling ``to_qc`` on *jaspr*.

    Iterates over the jaspr's invars and replaces ``QubitArray`` types with
    concrete qubit lists from an analysis circuit.  Static parameters have
    been bound via closure before tracing, so only QuantumArray-derived
    invars remain.
    """
    qubits = list(QC(n_qubits).qubits)

    # Partition qubits into slices corresponding to each QuantumArray arg
    off, slices = 0, []
    for a in quantum_arrays:
        n = a.size * a.qtype.size
        slices.append(qubits[off : off + n])
        off += n
    it = iter(slices)

    args = []
    for v in jaspr.invars[:-1]:  # exclude trailing QC
        s = str(v.aval)
        if s == "QubitArray":
            args.append(next(it))
        elif "int" in s and getattr(v.aval, "shape", ()):
            args.append(np.arange(n_qubits))  # ind_array for qubit indexing
        elif "int" in s or "bool" in s:
            args.append(1)  # qtype_size (always 1 for QuantumBool)
        else:
            raise ValueError(f"Unexpected invar: {s}")
    return args


def _prepare_for_tqecd(stim_circ):
    """Re-structure *stim_circ* into the fragment format expected by tqecd.

    tqecd requires each *fragment* to follow this layout:

    1. Zero or more moments of reset operations
    2. Zero or more moments of computation gates (no resets or measurements)
    3. Exactly one moment of measurement operations

    Moments are separated by ``TICK`` instructions.  Multi-round circuits are
    split at measurement boundaries so each round becomes its own fragment.
    The first fragment receives implicit ``R`` (Z-basis reset) for every qubit
    that was not explicitly reset by the user's circuit.
    """
    # Collect all qubit indices
    all_qubits = {
        t.value
        for i in stim_circ.flattened()
        for t in i.targets_copy()
        if not t.is_measurement_record_target
    }

    # Split instructions into rounds.
    # A round boundary occurs when we've accumulated measurements and the
    # next instruction is no longer a measurement.
    rounds, cur = [], {"R": [], "G": [], "M": []}
    for inst in stim_circ.flattened():
        if inst.name == "TICK":
            continue
        if cur["M"] and inst.name not in _MEASUREMENT_OPS:
            rounds.append(cur)
            cur = {"R": [], "G": [], "M": []}
        key = (
            "R"
            if inst.name in _RESET_OPS
            else "M" if inst.name in _MEASUREMENT_OPS else "G"
        )
        cur[key].append(inst)
    if cur["M"]:
        rounds.append(cur)

    out = stim.Circuit()
    for q in sorted(all_qubits):
        out.append("QUBIT_COORDS", [q], [float(q), 0.0])

    for ri, rnd in enumerate(rounds):
        # First round: implicit R for un-reset qubits (tqecd requires it)
        reset_qbs = {t.value for i in rnd["R"] for t in i.targets_copy()}
        if ri == 0 and (unreset := sorted(all_qubits - reset_qbs)):
            out.append("R", unreset)
        for i in rnd["R"]:
            out.append(i.name, _tvals(i), i.gate_args_copy())
        out.append("TICK")

        # Gate moments — two-qubit gates each get their own moment

        for i in rnd["G"]:
            tgts, ga = i.targets_copy(), i.gate_args_copy()
            if i.name in _TWO_QUBIT_OPS:
                for j in range(0, len(tgts), 2):
                    out.append(i.name, [tgts[j].value, tgts[j + 1].value], ga)
                    out.append("TICK")
            else:
                out.append(i.name, _tvals(i), ga)
                out.append("TICK")

        # Measurements
        for i in rnd["M"]:
            out.append(i.name, _tvals(i), i.gate_args_copy())
        if ri < len(rounds) - 1:
            out.append("TICK")
    return out


def _extract_detectors(annotated_circ, total_meas, inv_meas_map):
    """Extract detector definitions from an annotated stim circuit.

    Returns a list of lists of :class:`~qrisp.circuit.Clbit`, one list per
    ``DETECTOR`` instruction.
    """
    return [
        [
            inv_meas_map[total_meas + t.value]
            for t in inst.targets_copy()
            if t.is_measurement_record_target
        ]
        for inst in annotated_circ.flattened()
        if inst.name == "DETECTOR"
    ]


# ───────────────────── public API ──────────────────────────────────────


def find_detectors(func=None, *, return_circuits=False):
    r"""Decorator that automatically identifies stim detectors and returns them.

    ``find_detectors`` leverages `tqecd <https://github.com/tqec/tqecd>`_
    to automatically discover detector parity checks in quantum error correction
    circuits. The detected parities are returned as an additional value alongside
    the decorated function's original return values.

    .. note::

        This feature requires the optional ``tqecd`` package.  Install it
        as described `here <https://tqec.github.io/tqecd/index.html>`_.

    Pitfalls & Limitations
    ^^^^^^^^^^^^^^^^^^^^^^

    The ``find_detectors`` feature comes with some constraints:

    **QuantumArray arguments must be QuantumBool**
        All ``QuantumArray`` arguments to the decorated function must have
        ``qtype=QuantumBool()``.  Other quantum types will raise a
        ``TypeError``.  Static parameters (integers, booleans, etc.) are
        allowed and will be passed through to the function unchanged.

    **Only Clifford gates are supported**
        tqecd analyses stabilizer flows, which are only well-defined for
        Clifford circuits.  If the decorated function contains non-Clifford
        gates (e.g. T, Toffoli, arbitrary-angle rotations), stim will raise
        an error during circuit construction or tqecd will silently produce
        incorrect results.

    **No real-time classical computation inside the decorated function**
        The circuit is traced symbolically.  Conditional logic that depends
        on measurement outcomes (real-time ``if``/``else``) cannot be
        expressed in the stim circuit and will break analysis.

    **Composite detectors from flow products are not discoverable**
        tqecd finds detectors by tracking individual stabilizer flows from
        resets through gates to measurements.  It **cannot** discover
        composite detectors that arise as products of multiple stabilizer
        flows.  For example, in a GHZ-state circuit the parity
        :math:`Z_3 Z_4` is a product of :math:`Z_0 Z_3` and
        :math:`Z_0 Z_4`, but tqecd will not identify it.

    **Multi-round circuits require explicit resets**
        For tqecd to detect round-to-round detectors (comparing syndrome
        measurements across rounds), ancilla qubits **must** be explicitly
        reset between rounds.  Without resets, tqecd cannot establish the
        start of a new stabilizer flow and will miss cross-round detectors.

        This is critical: in a two-round repetition code without resets
        between rounds, tqecd will only find first-round detectors and
        miss the temporal detectors that compare round 1 vs round 2.

    **Detectors referencing non-returned measurements are discarded**
        If a detector involves a measurement whose result is *not* part of
        the decorated function's return value, the detector is silently
        dropped.  Make sure all relevant measurements are returned.

        For example, if you measure both syndrome qubits but only return
        one measurement, detectors involving the unreturned measurement
        will be filtered out.

    **Multiple QuantumArray arguments receive disjoint qubit slices**
        When the decorated function takes more than one QuantumArray
        argument, it is assumed that each argument indeed represents
        a distinct set of qubits. This can not be guaranteed at tracing
        time and needs to be ensured from user side.
        Each array is mapped to its own contiguous slice of
        qubits during analysis.  The slices are allocated in positional
        order.

    **May discover more detectors than expected**
        In some circuits, ``find_detectors`` may identify *more* valid
        detectors than a minimal set.  For example, in 2-round codes,
        tqecd can discover boundary detectors from data qubit measurements
        that are mathematically valid but redundant with syndrome-based
        detectors.  All discovered detectors are correct (they will sample
        to zero in noiseless circuits); the extra ones represent additional
        stabilizer flows through the circuit.

    Parameters
    ----------
    func : callable, optional
        The function to decorate.  When using keyword arguments (e.g.
        ``@find_detectors(return_circuits=True)``), *func* is ``None`` and
        the decorator returns a wrapper that accepts *func*.

        The decorated function must accept at least one ``QuantumArray``
        argument (with ``qtype=QuantumBool()``).  Additional non-QuantumArray
        arguments (integers, booleans, strings, etc.) are treated as static
        parameters: they are captured by closure before JAX tracing so they
        behave as compile-time constants inside the function body.
    return_circuits : bool, default ``False``
        When ``True``, three additional items are appended to the return
        value (after the detector list and the original returns):

        * **raw_stim** — the stim circuit obtained directly from ``to_qc``
          ``to_stim``, before any restructuring.
        * **tqecd_input** — the circuit after ``_prepare_for_tqecd`` has
          re-arranged it into proper fragments (the input fed to tqecd).
        * **annotated** — the circuit that comes back from
          ``tqecd.annotate_detectors_automatically``, containing the
          ``DETECTOR`` instructions.

        This is useful for debugging or understanding why detectors were
        or were not discovered.

    Returns
    -------
    The decorated function returns:
        (detector_bools, \*original_returns)
    or, when ``return_circuits=True``:
        (detector_bools, \*original_returns, raw_stim, tqecd_input, annotated)
        where *detector_bools* is a ``list`` of Jasp-traced boolean values, one
        per discovered detector, each representing the ``parity()`` of the
        measurements that constitute that detector. The original return values
        from the decorated function are unpacked and appended after the detector list.

    Examples
    --------
    **Example 1: Single-round syndrome extraction**

    Three data qubits with two ancilla parity checks::

        from qrisp import *
        from qrisp.jasp.evaluation_tools.stim_extraction import extract_stim
        from qrisp.misc.stim_tools import find_detectors

        @find_detectors
        def syndrome(data, ancilla):
            # Reset ancilla qubits
            reset(ancilla[0])
            reset(ancilla[1])

            # Syndrome extraction
            cx(data[0], ancilla[0])
            cx(data[1], ancilla[0])
            cx(data[1], ancilla[1])
            cx(data[2], ancilla[1])

            # Measure syndromes
            return measure(ancilla[0]), measure(ancilla[1])

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(3,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(2,))
            detectors, m0, m1 = syndrome(data, ancilla)
            return detectors, m0, m1

        # Run and extract stim circuit
        result = main()
        stim_circ = result[-1]
        print(f"Found {stim_circ.num_detectors} detectors")

    **Example 2: Two-round repetition code**

    Distance-3 repetition code with temporal detectors comparing rounds::

        @find_detectors
        def rep_code_2rounds(data, ancilla):
            measurements = []

            for round_num in range(2):
                # Reset ancillas
                reset(ancilla[0])
                reset(ancilla[1])

                # Parity checks
                cx(data[0], ancilla[0])
                cx(data[1], ancilla[0])
                cx(data[1], ancilla[1])
                cx(data[2], ancilla[1])

                # Measure syndromes
                measurements.append(measure(ancilla[0]))
                measurements.append(measure(ancilla[1]))

            # Final data readout
            for i in range(3):
                measurements.append(measure(data[i]))

            return tuple(measurements)

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(3,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(2,))
            detectors, *measurements = rep_code_2rounds(data, ancilla)
            return (detectors,) + tuple(measurements)

        result = main()
        stim_circ = result[-1]
        # Will find >= 6 detectors (may discover more valid boundary detectors):
        # - 2 detectors in round 1 (vs initial reset)
        # - 2 detectors in round 2 (vs round 1)
        # - 2+ data-boundary detectors (tqecd may find additional valid ones)
        print(f"Found {stim_circ.num_detectors} detectors")

    **Example 3: Debug mode with circuit inspection**

    Use ``return_circuits=True`` to inspect intermediate circuits::

        @find_detectors(return_circuits=True)
        def debug_syndrome(data, ancilla):
            reset(ancilla[0])
            cx(data[0], ancilla[0])
            cx(data[1], ancilla[0])
            return measure(ancilla[0])

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(2,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(1,))
            detectors, m, raw_stim, tqecd_input, annotated = debug_syndrome(data, ancilla)

            print("Raw stim circuit:")
            print(raw_stim)
            print("\nAfter tqecd restructuring:")
            print(tqecd_input)
            print("\nWith detectors annotated:")
            print(annotated)

            return detectors, m

        main()

    **Example 4: Multi-array arguments**

    Separate data and ancilla qubits::

        @find_detectors
        def surface_check(data_qubits, ancilla_qubits):
            # Reset ancilla
            reset(ancilla_qubits[0])

            # X-stabilizer (Hadamard + CNOTs)
            h(ancilla_qubits[0])
            cx(ancilla_qubits[0], data_qubits[0])
            cx(ancilla_qubits[0], data_qubits[1])
            cx(ancilla_qubits[0], data_qubits[2])
            cx(ancilla_qubits[0], data_qubits[3])
            h(ancilla_qubits[0])

            return measure(ancilla_qubits[0])

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(4,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(1,))
            detectors, m = surface_check(data, ancilla)
            return detectors, m

        result = main()
        # Note: This finds 0 detectors because the X-stabilizer measurement
        # on Z-basis |0⟩ initialized qubits is not deterministic.
        # For detectors, would need X-basis initialization (RX) on data qubits.

    **Example 5: Three-round code with explicit resets**

    Shows importance of reset between rounds::

        @find_detectors
        def three_rounds(data, ancilla):
            results = []

            for _ in range(3):
                # CRITICAL: Reset between rounds
                reset(ancilla[0])

                # Syndrome extraction
                cx(data[0], ancilla[0])
                cx(data[1], ancilla[0])

                results.append(measure(ancilla[0]))

            return tuple(results)

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(2,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(1,))
            detectors, m1, m2, m3 = three_rounds(data, ancilla)
            return detectors, m1, m2, m3

        result = main()
        stim_circ = result[-1]
        # Will find >= 3 detectors (one per round plus temporal)
        print(f"Found {stim_circ.num_detectors} detectors")

        # Verify all detectors are valid (noiseless samples = 0)
        sampler = stim_circ.compile_detector_sampler()
        samples = sampler.sample(shots=1000)
        assert samples.sum() == 0, "All detectors should be deterministic!"

    **Example 6: Partial measurement return**

    Only detectors involving returned measurements are kept::

        @find_detectors
        def partial_return(data, ancilla):
            reset(ancilla[0])
            reset(ancilla[1])

            cx(data[0], ancilla[0])
            cx(data[1], ancilla[0])
            cx(data[1], ancilla[1])
            cx(data[2], ancilla[1])

            m0 = measure(ancilla[0])
            m1 = measure(ancilla[1])

            # Only return m0, discard m1
            return m0

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(3,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(2,))
            detectors, m = partial_return(data, ancilla)
            return detectors, m

        result = main()
        stim_circ = result[-1]
        # Only detectors involving m0 (the returned measurement) are kept.
        # Detectors that depend on m1 (unreturned) are filtered out.
        # Typically finds 1 detector from the m0 measurement vs initial reset.
        print(f"Found {stim_circ.num_detectors} detectors")

    **Example 7: Static parameters (integers, booleans)**

    The decorated function can accept non-QuantumArray arguments such as
    integers or booleans.  These are bound via closure before tracing so
    that JAX treats them as compile-time constants::

        @find_detectors
        def configurable_syndrome(data, ancilla, num_rounds, use_extra_gate):
            results = []
            for _ in range(num_rounds):
                reset(ancilla[0])
                cx(data[0], ancilla[0])
                cx(data[1], ancilla[0])
                if use_extra_gate:
                    cx(data[1], ancilla[0])
                    cx(data[1], ancilla[0])  # cancels out
                results.append(measure(ancilla[0]))
            return tuple(results)

        @extract_stim
        def main():
            data = QuantumArray(qtype=QuantumBool(), shape=(2,))
            ancilla = QuantumArray(qtype=QuantumBool(), shape=(1,))
            detectors, *ms = configurable_syndrome(data, ancilla, 3, True)
            return (detectors,) + tuple(ms)

        result = main()
        stim_circ = result[-1]
        print(f"Found {stim_circ.num_detectors} detectors")
    """
    try:
        import tqecd
    except ImportError:
        raise ImportError(
            "find_detectors requires the 'tqecd' package. "
            "Install it with: pip install tqecd"
        )

    def _decorator(fn):
        def wrapper(*args, **kwargs):
            # --- Validate & classify arguments ---
            quantum_arrays = []
            for a in args:
                if isinstance(a, QuantumArray):
                    if not isinstance(a.qtype, QuantumBool):
                        raise TypeError(
                            f"find_detectors: QuantumArray args must have "
                            f"qtype QuantumBool, got {a.qtype}"
                        )
                    quantum_arrays.append(a)
            if not quantum_arrays:
                raise TypeError(
                    "find_detectors: at least one QuantumArray argument is required"
                )

            n_qubits = sum(a.size * a.qtype.size for a in quantum_arrays)

            # --- 1. Trace inner function to get sub-jaspr ---
            # Bind non-QuantumArray args via closure so JAX only traces
            # the quantum arguments.  Static values (ints, bools, …)
            # become constants inside the jaspr.
            qa_idx = [i for i, a in enumerate(args) if isinstance(a, QuantumArray)]
            static = {
                i: a for i, a in enumerate(args) if not isinstance(a, QuantumArray)
            }

            def fn_qa_only(*qa_args, _kw=kwargs):
                full = [None] * len(args)
                for i, q in zip(qa_idx, qa_args):
                    full[i] = q
                for i, v in static.items():
                    full[i] = v
                return fn(*full, **_kw)

            qa_args = tuple(args[i] for i in qa_idx)
            # Count the expected number of flat leaves from the
            # QuantumArray args.  JAX's make_jaxpr creates one invar per
            # leaf plus a trailing QC invar appended by Qrisp.
            expected_n_invars = len(tree_flatten(qa_args)[0]) + 1  # +1 for QC

            _TRACER_MSG = (
                "find_detectors: a JAX tracer was leaked into the "
                "decorated function via a non-QuantumArray argument.  "
                "All non-QuantumArray parameters (including values "
                "nested inside dicts, lists, etc.) must be concrete "
                "Python values, not traced quantities."
            )

            try:
                jaspr = make_jaspr(fn_qa_only)(*qa_args)
            except (
                jax.errors.UnexpectedTracerError,
                jax.errors.TracerIntegerConversionError,
                TypeError,
            ) as exc:
                raise TypeError(_TRACER_MSG) from exc

            # --- 2. Analysis: to_qc → stim → tqecd ---
            result = jaspr.to_qc(*_build_analysis_args(jaspr, n_qubits, quantum_arrays))

            # Flatten analysis Clbits (consistent ordering via tree_flatten)
            analysis_clbits, _ = tree_flatten(result[:-1])

            # Convert to stim
            raw_stim, meas_map = result[-1].to_stim(return_measurement_map=True)
            inv_map = {v: k for k, v in meas_map.items()}

            # Prepare for tqecd and annotate
            tqecd_circ = _prepare_for_tqecd(raw_stim)
            annotated = tqecd.annotate_detectors_automatically(tqecd_circ)

            # Extract detector Clbit sets
            all_dets = _extract_detectors(
                annotated, tqecd_circ.num_measurements, inv_map
            )

            # Keep only detectors whose Clbits all appear in returned measurements
            returned = set(analysis_clbits)
            relevant = [
                (i, d)
                for i, d in enumerate(all_dets)
                if all(cb in returned for cb in d)
            ]

            # --- 2b. Simulate noiseless circuit to determine detector expectations ---
            # Use stim's built-in method to remove all noise processes
            expectations = (
                annotated.without_noise()
                .compile_detector_sampler()
                .sample(shots=1)[0]
                .tolist()
            )

            # --- 3. Call real function & map analysis Clbits to traced booleans ---
            real_returns = fn(*args, **kwargs)
            if not isinstance(real_returns, tuple):
                real_returns = (real_returns,)
            traced_flat, _ = tree_flatten(real_returns)
            cb_to_traced = dict(zip(analysis_clbits, traced_flat))

            # --- 4. Emit parity calls for each detector ---
            det_results = [
                parity(
                    *[cb_to_traced[cb] for cb in det],
                    expectation=int(expectations[idx]),
                )
                for idx, det in relevant
            ]

            out = (det_results,) + real_returns
            if return_circuits:
                out += (raw_stim, tqecd_circ, annotated)
            return out

        return wrapper

    # Support both @find_detectors and @find_detectors(return_circuits=True)
    if func is not None:
        return _decorator(func)
    return _decorator



"""
********************************************************************************
* Copyright (c) 2026 the Qrisp authors
*
* This program and the accompanying materials are made available under the
* terms of the Eclipse Public License 2.0 which is available at
* http://www.eclipse.org/legal/epl-2.0.
*
* This Source Code may also be made available under the following Secondary
* Licenses when the conditions for such availability set forth in the Eclipse
* Public License, v. 2.0 are satisfied: GNU General Public License, version 2
* with the GNU Classpath Exception which is
* available at https://www.gnu.org/software/classpath/license.html.
*
* SPDX-License-Identifier: EPL-2.0 OR GPL-2.0 WITH Classpath-exception-2.0
********************************************************************************
"""

import numpy as np
import sympy as sp
from qrisp.core import QuantumVariable
from qrisp.operators.qubit import X, Y, Z
from qrisp.algorithms.cold import DCQOProblem, solve_alpha_gamma_chi, solve_alpha


def create_COLD_instance(Q, uniform_AGP_coeffs):
    """
    Create the necessary parameters and operators to initialize a DCQO problem instance for COLD.

    Parameters
    ----------
    Q : np.array
        The QUBO Matrix to be encoded in the Hamiltonian.
    uniform_AGP_coeffs : bool
        Whether to approximate the AGP with uniform or non-uniform coefficients.

    Returns
    -------
    collected operators : tuple
        Tuple containing the following functions and operators:
        Scheduling function (lam(t)), function for AGP coefficients (alpha), initial and problem
        Hamiltonian (H_init, H_prob), AGP function (A_lam), Coupling and onsite energies of
        problem Hamiltonian (J, h), inverse scheduling function (g(lam)), control Hamiltonian (H_control).
    """

    N = len(Q[0])
    h = -0.5 * np.diag(Q) - 0.5 * np.sum(Q, axis=1)
    J = 0.5 * Q

    def lam():
        t, T = sp.symbols("t T", real=True)
        lam_expr = t / T
        return lam_expr

    def g():
        # Inverse of lam(t) giving t(lam)
        lam, T = sp.symbols("lam T")
        g_expr = lam * T
        return g_expr

    # AGP coefficients
    if uniform_AGP_coeffs:

        def alpha(lam, f, f_deriv):
            A = lam * h + f
            B = 1 - lam
            C = h + f_deriv

            nom = np.sum(A + 4 * B * C)
            denom = 2 * (np.sum(A**2) + N * (B**2)) + 4 * (lam**2) * np.sum(
                np.tril(J, -1).sum(axis=1)
            )
            alph = nom / denom
            alph = [alph] * N

            return alph

    else:

        def alpha(lam, f, f_deriv):
            nom = [h[i] + f + (1 - lam) * f_deriv for i in range(N)]
            denom = [
                2
                * (
                    (lam * h[i] + f) ** 2
                    + (1 - lam) ** 2
                    + lam**2 * sum([J[i][j] for j in range(N) if j != i])
                )
                for i in range(N)
            ]

            alph = [nom[i] / denom[i] for i in range(N)]
            return alph

    # Initial Hamiltonian
    H_init = 1 * sum([X(i) for i in range(N)])

    # Problem Hamiltonian
    H_prob = sum(
        [sum([J[i][j] * Z(i) * Z(j) for j in range(i)]) for i in range(N)]
    ) + sum([h[i] * Z(i) for i in range(N)])

    # AGP as function of alpha
    if uniform_AGP_coeffs:
        A_lam = sum([Y(i) for i in range(N)])
    else:
        A_lam = [Y(i) for i in range(N)]

    # Control Hamiltonian
    H_control = sum([Z(i) for i in range(N)])

    collected_operators = (Q, H_init, H_prob, A_lam, alpha, lam, g, H_control)

    return collected_operators


def create_LCD_instance(Q, agp_type, uniform_AGP_coeffs=True):
    """
    Create the necessary parameters and operators to initialize a DCQO problem instance for LCD.

    Parameters
    ----------
    Q : np.array
        The QUBO Matrix to be encoded in the Hamiltonian.
    agp_type : str
        Which approximation of the AGP to use. Can choose between ``order1``,
        ``order2``, ``nc`` (nested commutators up to first order).
    uniform_AGP_coeffs : bool
        Whether to approximate the AGP with uniform or non-uniform coefficients.

    Returns
    -------
    collected operators : tuple
        Tuple containing the following functions and operators:
        Scheduling function (lam(t)), function for AGP coefficients (alpha), initial and problem
        Hamiltonian (H_init, H_prob), AGP function (A_lam), Coupling and onsite energies of
        problem Hamiltonian (J, h), inverse scheduling function (g(lam)), control Hamiltonian (H_control).
    """

    def build_agp(agp_type, J, h):

        def order1():
            A_lam = [Y(i) for i in range(N)]
            return A_lam

        def nested_commutators(J, h):
            A_lam = -2 * [
                h[i] * Y(i)
                + sum([J[i][j] * (Y(i) * Z(j) + Z(i) * Y(j)) for j in range(i)])
                for i in range(N)
            ]
            return A_lam

        builders = {"order1": order1(), "nc": nested_commutators(J, h)}

        return builders[agp_type]

    def build_coeffs(agp_type, uniform_AGP_coeffs, J, h):

        def order1_uniform(J, h):
            def alpha(lam):
                A = lam * h
                B = 1 - lam
                nom = np.sum(A + 4 * B * h)
                denom = 2 * (np.sum(A**2) + N * (B**2)) + 4 * (lam**2) * np.sum(
                    np.tril(J, -1).sum(axis=1)
                )
                alph = nom / denom
                alph = [alph] * N
                return alph

            return alpha

        def order1_nonuniform(J, h):
            def alpha(lam):
                denom = [
                    2
                    * (
                        (lam * h[i]) ** 2
                        + (1 - lam) ** 2
                        + lam**2 * sum([J[i][j] for j in range(N) if j != i])
                    )
                    for i in range(N)
                ]
                alph = [h[i] / denom[i] for i in range(N)]
                return alph

            return alpha

        def nc_uniform(J, h):
            def alpha(lam):
                S_hR = sum(
                    [
                        sum([J[i][j] ** 2 * (h[i] + h[j]) for i in range(j)])
                        for j in range(N)
                    ]
                )
                S_hsqR = sum(
                    [
                        sum([J[i][j] ** 2 * (h[i] ** 2 + h[j] ** 2) for i in range(j)])
                        for j in range(N)
                    ]
                )
                S_2 = sum([sum([J[i][j] ** 2 for i in range(j)]) for j in range(N)])
                S_4 = sum([sum([J[i][j] ** 4 for i in range(j)]) for j in range(N)])
                R_i_list = [
                    sum([J[i][j] ** 2 if j != i else 0 for j in range(N)])
                    for i in range(N)
                ]
                S_Rsq = sum(R_i**2 for R_i in R_i_list)
                S_h = sum(h)
                S_hsq = sum(i**2 for i in h)

                nom = S_h + 2 * S_2
                denom = 4 * (
                    lam**2
                    * (S_hsq + 2 * S_hsqR + 6 * S_hR + 2 * S_Rsq + 4 * S_2 - 2 * S_4)
                    + (1 - lam) ** 2 * (N + 8 * S_2)
                )

                alph = -nom / denom
                alph = [N * [alph]]
                return alph

            return alpha

        def nc_nonuniform(J, h):
            def alpha(lam):
                alph = [solve_alpha(h, J, lam)]
                return alph

            return alpha

        builders = {
            ("order1", True): order1_uniform(J, h),
            ("order1", False): order1_nonuniform(J, h),
            ("nc", True): nc_uniform(J, h),
            ("nc", False): nc_nonuniform(J, h),
        }

        return builders[(agp_type, uniform_AGP_coeffs)]

    N = len(Q[0])
    h = -0.5 * np.diag(Q) - 0.5 * np.sum(Q, axis=1)
    J = 0.5 * Q

    def lam():
        t, T = sp.symbols("t T", real=True)
        lam_expr = sp.sin(sp.pi / 2 * sp.sin(sp.pi * t / (2 * T)) ** 2) ** 2
        return lam_expr

    # AGP coefficients
    coeff_func = build_coeffs(agp_type, uniform_AGP_coeffs, J, h)

    # Initial Hamiltonian
    H_init = 1 * sum([X(i) for i in range(N)])

    # Problem Hamiltonian
    H_prob = sum(
        [sum([J[i][j] * Z(i) * Z(j) for j in range(i)]) for i in range(N)]
    ) + sum([h[i] * Z(i) for i in range(N)])

    # AGP
    A_lam = build_agp(agp_type, J, h)

    return Q, H_init, H_prob, A_lam, coeff_func, lam


def solve_QUBO(Q: np.array, problem_args: dict, run_args: dict):
    """
    Solves a QUBO Matrix using counterdiabatic driving. This method uses the pre-defined COLD/LCD operators
    (hamiltonian, scheduling function, AGP parameters) as described in the tutorial.
    To define your own operators, create a DCQO instance and use the ``run`` method.

    Parameters
    ----------

    Q : np.array
        QUBO Matrix to solve.
    problem_args : dict
        Holds arguments for DCQO problem creation (``method``: str ("COLD"/"LCD"), ``uniform``: bool).
    run_args : dict
        Holds arguments for running the DCQO instance (``N_steps``, ``T``, ``N_opt``, ``CRAB``).
        For all options, see :meth:`DCQOProblem.run`.

    Returns
    -------

    result : dict
        The dictionary holding the QUBO vector results, with their probabilitites and cost.
        They are ordered from most to least likely and the dictionary entries are {"state": [prob, cost]}.


    Examples
    --------

    ::

        import numpy as np
        from qrisp.algorithms.cold import solve_QUBO

        Q = np.array([[-1.1, 0.6, 0.4, 0.0, 0.0, 0.0],
                    [0.6, -0.9,  0.5, 0.0, 0.0, 0.0],
                    [0.4, 0.5, -1.0, -0.6, 0.0, 0.0],
                    [0.0, 0.0, -0.6, -0.5, 0.6, 0.0],
                    [0.0, 0.0, 0.0, 0.6, -0.3, 0.5],
                    [0.0, 0.0, 0.0, 0.0, 0.5, -0.4]])

        problem_args = {"method": "COLD", "uniform": False}

        run_args = {"N_steps": 4, "T": 8, "N_opt": 1, "CRAB": False,
                    "objective": "agp_coeff_magnitude", "bounds": (-3, 3)}

        result = solve_QUBO(Q, problem_args, run_args)

        print(result)

    ::

        {'101101': [0.2749327493274933, np.float64(-3.4)], '011101': [0.14747147471474714, np.float64(-3.0)], '111101': [0.10500105001050011, np.float64(-2.1)], '101100': [0.0888608886088861, np.float64(-3.0)], '011100': [0.04720047200472005, np.float64(-2.6)], '101110': [0.043830438304383046, np.float64(-2.1)], '111100': [0.035760357603576036, np.float64(-1.7000000000000002)], '011110': [0.024410244102441025, np.float64(-1.7)], '100101': [0.02137021370213702, np.float64(-2.0)], '110101': [0.016960169601696017, np.float64(-1.7000000000000002)], '111110': [0.01686016860168602, np.float64(-0.8)], '111001': [0.016390163901639016, np.float64(-0.4)], '001101': [0.014560145601456015, np.float64(-3.1)], '010101': [0.014370143701437016, np.float64(-1.7999999999999998)], '101011': [0.011320113201132012, np.float64(-1.0)], '111000': [0.008620086200862008, np.float64(0.0)], '100001': [0.007290072900729008, np.float64(-1.5)], '000101': [0.006580065800658007, np.float64(-0.9)], '110100': [0.006420064200642007, np.float64(-1.3000000000000003)], '011011': [0.006090060900609006, np.float64(-0.6)], '101000': [0.005780057800578007, np.float64(-1.3)], '100100': [0.0057700577005770064, np.float64(-1.6)], '101001': [0.005720057200572007, np.float64(-1.7000000000000002)], '101111': [0.005310053100531005, np.float64(-1.5)], '100110': [0.005230052300523006, np.float64(-0.7)], '010100': [0.004050040500405004, np.float64(-1.4)], '001100': [0.004040040400404004, np.float64(-2.7)], '011000': [0.003870038700387004, np.float64(-0.9)], '011111': [0.0032400324003240034, np.float64(-1.1)], '100000': [0.0031800318003180035, np.float64(-1.1)], '110010': [0.0029900299002990033, np.float64(-1.1)], '010001': [0.002930029300293003, np.float64(-1.3)], '011001': [0.002850028500285003, np.float64(-1.3)], '110011': [0.002770027700277003, np.float64(-0.5000000000000001)], '001110': [0.002650026500265003, np.float64(-1.8)], '010110': [0.0025900259002590025, np.float64(-0.5)], '110001': [0.0024400244002440027, np.float64(-1.2000000000000002)], '000100': [0.0022800228002280024, np.float64(-0.5)], '100011': [0.002020020200202002, np.float64(-0.8000000000000002)], '111011': [0.001770017700177002, np.float64(0.3)], '010000': [0.0015300153001530014, np.float64(-0.9)], '101010': [0.0014900149001490016, np.float64(-1.6)], '111111': [0.0014200142001420015, np.float64(-0.20000000000000007)], '000010': [0.0010500105001050011, np.float64(-0.3)], '110111': [0.0008300083000830009, np.float64(0.19999999999999984)], '010011': [0.0008200082000820009, np.float64(-0.6)], '001011': [0.0007900079000790008, np.float64(-0.7000000000000001)], '100111': [0.0007200072000720008, np.float64(-0.09999999999999998)], '001000': [0.0006300063000630007, np.float64(-1.0)], '100010': [0.0006000060000600005, np.float64(-1.4000000000000001)], '001111': [0.0005800058000580006, np.float64(-1.2000000000000002)], '010010': [0.0005700057000570006, np.float64(-1.2)], '001001': [0.0005300053000530005, np.float64(-1.4)], '011010': [0.0005200052000520005, np.float64(-1.2)], '000011': [0.0005000050000500006, np.float64(0.3)], '111010': [0.0004300043000430004, np.float64(-0.3)], '110000': [0.00026000260002600025, np.float64(-0.8000000000000002)], '000110': [0.00026000260002600025, np.float64(0.39999999999999997)], '010111': [0.00021000210002100023, np.float64(0.09999999999999998)], '000001': [0.0001700017000170002, np.float64(-0.4)], '001010': [0.00012000120001200013, np.float64(-1.3)], '000000': [9.00009000090001e-05, np.float64(0.0)], '000111': [8.000080000800009e-05, np.float64(1.0)], '110110': [2.0000200002000023e-05, np.float64(-0.4000000000000002)]}

    """

    method = problem_args["method"]

    if method == "LCD":
        # Check if AGP type is specified, otherwise use 1st order
        try:
            agp_type = problem_args["agp_type"]
        except KeyError:
            agp_type = "order1"

        problem_operators = create_LCD_instance(
            Q, agp_type=agp_type, uniform_AGP_coeffs=problem_args["uniform"]
        )

    elif method == "COLD":
        problem_operators = create_COLD_instance(
            Q, uniform_AGP_coeffs=problem_args["uniform"]
        )

    # Create qarg and problem instrance
    qarg = QuantumVariable(Q.shape[0])
    prob = DCQOProblem(*problem_operators)

    # Run problem
    result = prob.run(qarg, method=method, **run_args)

    return result
